@@ -1,11 +1,27 @@
-# -*- test-case-name: buildbot.test.test_steps,buildbot.test.test_properties -*-
+# This file is part of Buildbot.  Buildbot is free software: you can
+# redistribute it and/or modify it under the terms of the GNU General Public
+# License as published by the Free Software Foundation, version 2.
+#
+# This program is distributed in the hope that it will be useful, but WITHOUT
+# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
+# details.
+#
+# You should have received a copy of the GNU General Public License along with
+# this program; if not, write to the Free Software Foundation, Inc., 51
+# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
+#
+# Copyright Buildbot Team Members
+
 
 import re
 from twisted.python import log
 from twisted.spread import pb
 from buildbot.process.buildstep import LoggingBuildStep, RemoteShellCommand
 from buildbot.process.buildstep import RemoteCommand
-from buildbot.status.builder import SUCCESS, WARNINGS, FAILURE, STDOUT, STDERR
+from buildbot.status.results import SUCCESS, WARNINGS, FAILURE
+from buildbot.status.logfile import STDOUT, STDERR
+from buildbot.interfaces import BuildSlaveTooOldError
 
 # for existing configurations that import WithProperties from here.  We like
 # to move this class around just to keep our readers guessing.
@@ -50,6 +66,7 @@ class ShellCommand(LoggingBuildStep):
     """
 
     name = "shell"
+    renderables = [ 'description', 'descriptionDone', 'slaveEnvironment', 'remote_kwargs', 'command' ]
     description = None # set this to a list of short strings to override
     descriptionDone = None # alternate description when the step is complete
     command = None # set this to a command, or set in kwargs
@@ -101,6 +118,14 @@ class ShellCommand(LoggingBuildStep):
         # we need to stash the RemoteShellCommand's args too
         self.addFactoryArguments(**kwargs)
 
+    def setBuild(self, build):
+        LoggingBuildStep.setBuild(self, build)
+        # Set this here, so it gets rendered when we start the step
+        self.slaveEnvironment = self.build.slaveEnvironment
+
+    def setStepStatus(self, step_status):
+        LoggingBuildStep.setStepStatus(self, step_status)
+
     def setDefaultWorkdir(self, workdir):
         rkw = self.remote_kwargs
         rkw['workdir'] = rkw['workdir'] or workdir
@@ -125,24 +150,33 @@ class ShellCommand(LoggingBuildStep):
                      'tests' ...)
         """
 
-        if done and self.descriptionDone is not None:
-            return list(self.descriptionDone)
-        if self.description is not None:
-            return list(self.description)
+        try:
+            if done and self.descriptionDone is not None:
+                return self.descriptionDone
+            if self.description is not None:
+                return self.description
 
-        properties = self.build.getProperties()
-        words = self.command
-        if isinstance(words, (str, unicode)):
-            words = words.split()
-        # render() each word to handle WithProperties objects
-        words = properties.render(words)
-        if len(words) < 1:
+            # we may have no command if this is a step that sets its command
+            # name late in the game (e.g., in start())
+            if not self.command:
+                return ["???"]
+
+            words = self.command
+            if isinstance(words, (str, unicode)):
+                words = words.split()
+            # render() each word to handle WithProperties objects
+            words = self.build.render(words)
+            if len(words) < 1:
+                return ["???"]
+            if len(words) == 1:
+                return ["'%s'" % words[0]]
+            if len(words) == 2:
+                return ["'%s" % words[0], "%s'" % words[1]]
+            return ["'%s" % words[0], "%s" % words[1], "...'"]
+        except:
+            log.msg("Error describing step")
+            log.err()
             return ["???"]
-        if len(words) == 1:
-            return ["'%s'" % words[0]]
-        if len(words) == 2:
-            return ["'%s" % words[0], "%s'" % words[1]]
-        return ["'%s" % words[0], "%s" % words[1], "...'"]
 
     def setupEnvironment(self, cmd):
         # merge in anything from Build.slaveEnvironment
@@ -151,14 +185,13 @@ class ShellCommand(LoggingBuildStep):
         # BuildProperties.
         # Environment variables passed in by a BuildStep override
         # those passed in at the Builder level.
-        properties = self.build.getProperties()
-        slaveEnv = self.build.slaveEnvironment
+        slaveEnv = self.slaveEnvironment
         if slaveEnv:
             if cmd.args['env'] is None:
                 cmd.args['env'] = {}
             fullSlaveEnv = slaveEnv.copy()
             fullSlaveEnv.update(cmd.args['env'])
-            cmd.args['env'] = properties.render(fullSlaveEnv)
+            cmd.args['env'] = fullSlaveEnv
             # note that each RemoteShellCommand gets its own copy of the
             # dictionary, so we shouldn't be affecting anyone but ourselves.
 
@@ -193,18 +226,17 @@ class ShellCommand(LoggingBuildStep):
         # this block is specific to ShellCommands. subclasses that don't need
         # to set up an argv array, an environment, or extra logfiles= (like
         # the Source subclasses) can just skip straight to startCommand()
-        properties = self.build.getProperties()
 
         warnings = []
 
         # create the actual RemoteShellCommand instance now
-        kwargs = properties.render(self.remote_kwargs)
-        kwargs['command'] = properties.render(self.command)
+        kwargs = self.remote_kwargs
+        command = self.command
+        kwargs['command'] = command
         kwargs['logfiles'] = self.logfiles
 
         # check for the usePTY flag
         if kwargs.has_key('usePTY') and kwargs['usePTY'] != 'slave-config':
-            slavever = self.slaveVersion("shell", "old")
             if self.slaveVersionIsOlderThan("svn", "2.7"):
                 warnings.append("NOTE: slave does not allow master to override usePTY\n")
 
@@ -219,8 +251,13 @@ class ShellCommand(LoggingBuildStep):
 class TreeSize(ShellCommand):
     name = "treesize"
     command = ["du", "-s", "-k", "."]
+    description = "measuring tree size"
+    descriptionDone = "tree size measured"
     kib = None
 
+    def __init__(self, *args, **kwargs):
+        ShellCommand.__init__(self, *args, **kwargs)
+
     def commandComplete(self, cmd):
         out = cmd.logs['stdio'].getText()
         m = re.search(r'^(\d+)', out)
@@ -242,21 +279,15 @@ class TreeSize(ShellCommand):
 
 class SetProperty(ShellCommand):
     name = "setproperty"
+    renderables = [ 'property' ]
 
-    def __init__(self, **kwargs):
-        self.property = None
-        self.extract_fn = None
-        self.strip = True
+    def __init__(self, property=None, extract_fn=None, strip=True, **kwargs):
+        self.property = property
+        self.extract_fn = extract_fn
+        self.strip = strip
 
-        if kwargs.has_key('property'):
-            self.property = kwargs['property']
-            del kwargs['property']
-        if kwargs.has_key('extract_fn'):
-            self.extract_fn = kwargs['extract_fn']
-            del kwargs['extract_fn']
-        if kwargs.has_key('strip'):
-            self.strip = kwargs['strip']
-            del kwargs['strip']
+        assert (property is not None) ^ (extract_fn is not None), \
+                "Exactly one of property and extract_fn must be set"
 
         ShellCommand.__init__(self, **kwargs)
 
@@ -264,16 +295,13 @@ class SetProperty(ShellCommand):
         self.addFactoryArguments(extract_fn=self.extract_fn)
         self.addFactoryArguments(strip=self.strip)
 
-        assert self.property or self.extract_fn, \
-            "SetProperty step needs either property= or extract_fn="
-
         self.property_changes = {}
 
     def commandComplete(self, cmd):
         if self.property:
             result = cmd.logs['stdio'].getText()
             if self.strip: result = result.strip()
-            propname = self.build.getProperties().render(self.property)
+            propname = self.property
             self.setProperty(propname, result, "SetProperty Step")
             self.property_changes[propname] = result
         else:
@@ -330,6 +358,8 @@ class SilentRemoteCommand(RemoteCommand):
         pass
 
 class WarningCountingShellCommand(ShellCommand):
+    renderables = [ 'suppressionFile' ]
+
     warnCount = 0
     warningPattern = '.*warning[: ].*'
     # The defaults work for GNU Make.
@@ -341,7 +371,7 @@ class WarningCountingShellCommand(ShellCommand):
     suppressionLineRe = re.compile(r"^\s*(.+?)\s*:\s*(.+?)\s*(?:[:]\s*([0-9]+)(?:-([0-9]+))?\s*)?$")
 
     def __init__(self, workdir=None,
-                 warningPattern=None, warningExtractor=None,
+                 warningPattern=None, warningExtractor=None, maxWarnCount=None,
                  directoryEnterPattern=None, directoryLeavePattern=None,
                  suppressionFile=None, **kwargs):
         self.workdir = workdir
@@ -360,6 +390,7 @@ class WarningCountingShellCommand(ShellCommand):
             self.warningExtractor = warningExtractor
         else:
             self.warningExtractor = WarningCountingShellCommand.warnExtractWholeLine
+        self.maxWarnCount = maxWarnCount
 
         # And upcall to let the base class do its work
         ShellCommand.__init__(self, workdir=workdir, **kwargs)
@@ -368,6 +399,7 @@ class WarningCountingShellCommand(ShellCommand):
                                  directoryEnterPattern=directoryEnterPattern,
                                  directoryLeavePattern=directoryLeavePattern,
                                  warningExtractor=warningExtractor,
+                                 maxWarnCount=maxWarnCount,
                                  suppressionFile=suppressionFile)
         self.suppressions = []
         self.directoryStack = []
@@ -435,9 +467,8 @@ class WarningCountingShellCommand(ShellCommand):
             for fileRe, warnRe, start, end in self.suppressions:
                 if ( (file == None or fileRe == None or fileRe.search(file)) and
                      (warnRe == None or  warnRe.search(text)) and
-                     lineNo != None and
-                     (start == None or start <= lineNo) and
-                     (end == None or end >= lineNo) ):
+                     ((start == None and end == None) or
+                      (lineNo != None and start <= lineNo and end >= lineNo)) ):
                     return
 
         warnings.append(line)
@@ -454,10 +485,8 @@ class WarningCountingShellCommand(ShellCommand):
 
         self.myFileWriter = StringFileWriter()
 
-        properties = self.build.getProperties()
-
         args = {
-            'slavesrc': properties.render(self.suppressionFile),
+            'slavesrc': self.suppressionFile,
             'workdir': self.workdir,
             'writer': self.myFileWriter,
             'maxsize': None,
@@ -491,6 +520,12 @@ class WarningCountingShellCommand(ShellCommand):
         return ShellCommand.start(self)
 
     def createSummary(self, log):
+        """
+        Match log lines against warningPattern.
+
+        Warnings are collected into another log for this step, and the
+        build-wide 'warnings-count' is updated."""
+
         self.warnCount = 0
 
         # Now compile a regular expression from whichever warning pattern we're
@@ -533,7 +568,8 @@ class WarningCountingShellCommand(ShellCommand):
         # If there were any warnings, make the log if lines with warnings
         # available
         if self.warnCount:
-            self.addCompleteLog("warnings", "\n".join(warnings) + "\n")
+            self.addCompleteLog("warnings (%d)" % self.warnCount,
+                    "\n".join(warnings) + "\n")
 
         warnings_stat = self.step_status.getStatistic('warnings', 0)
         self.step_status.setStatistic('warnings', warnings_stat + self.warnCount)
@@ -546,7 +582,8 @@ class WarningCountingShellCommand(ShellCommand):
 
 
     def evaluateCommand(self, cmd):
-        if cmd.rc != 0:
+        if ( cmd.rc != 0 or
+           ( self.maxWarnCount != None and self.warnCount > self.maxWarnCount ) ):
             return FAILURE
         if self.warnCount:
             return WARNINGS
@@ -628,12 +665,13 @@ class PerlModuleTest(Test):
         total = 0
         passed = 0
         failed = 0
-        rc = cmd.rc
+        rc = SUCCESS
+        if cmd.rc > 0:
+            rc = FAILURE
 
         # New version of Test::Harness?
-        try:
+        if "Test Summary Report" in lines:
             test_summary_report_index = lines.index("Test Summary Report")
-
             del lines[0:test_summary_report_index + 2]
 
             re_test_result = re.compile("^Result: (PASS|FAIL)$|Tests: \d+ Failed: (\d+)\)|Files=\d+, Tests=(\d+)")
@@ -642,16 +680,15 @@ class PerlModuleTest(Test):
             test_result_lines = [mo.groups() for mo in mos if mo]
 
             for line in test_result_lines:
-                if line[0] == 'PASS':
-                    rc = SUCCESS
-                elif line[0] == 'FAIL':
+                if line[0] == 'FAIL':
                     rc = FAILURE
-                elif line[1]:
+
+                if line[1]:
                     failed += int(line[1])
-                elif line[2]:
+                if line[2]:
                     total = int(line[2])
 
-        except ValueError: # Nope, it's the old version
+        else: # Nope, it's the old version
             re_test_result = re.compile("^(All tests successful)|(\d+)/(\d+) subtests failed|Files=\d+, Tests=(\d+),")
 
             mos = map(lambda line: re_test_result.search(line), lines)
@@ -667,7 +704,6 @@ class PerlModuleTest(Test):
 
                     test_totals_line = test_result_lines[1]
                     total_str = test_totals_line[3]
-                    rc = SUCCESS
                 else:
                     failed_str = test_result_line[1]
                     failed = int(failed_str)
